{
 "cells": [
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a364481e",
   "metadata": {},
   "outputs": [],
   "source": [
    "import time\n",
    "import random\n",
    "\n",
    "import numpy as np\n",
    "from numpy import random \n",
    "import pywt\n",
    "from skimage import data\n",
    "\n",
    "import os\n",
    "\n",
    "import scipy\n",
    "import scipy.io\n",
    "import imageio\n",
    "import matplotlib.pyplot as plt\n",
    "import matplotlib.image as mpimg\n",
    "\n",
    "\n",
    "# source https://matthieutrs.github.io/\n",
    "\n",
    "def coef2vec(coef, Nx, Ny):\n",
    "    \"\"\"\n",
    "    Convert wavelet coefficients to an array-type vector, inverse operation of vec2coef.\n",
    "    The initial wavelet coefficients are stocked in a list as follows:\n",
    "        [cAn, (cHm, cVn, cDn), ..., (cH1, cV1, cD1)],\n",
    "    and each element is a 2D array.\n",
    "    After the conversion, the returned vector is as follows:\n",
    "    [cAn.flatten(), cHn.flatten(), cVn.flatten(), cDn.flatten(), ...,cH1.flatten(), cV1.flatten(), cD1.flatten()].\n",
    "    \"\"\"\n",
    "    vec = []\n",
    "    bookkeeping = []\n",
    "    for ele in coef:\n",
    "        if type(ele) == tuple:\n",
    "            bookkeeping.append((np.shape(ele[0])))\n",
    "            for wavcoef in ele:\n",
    "                vec = np.append(vec, wavcoef.flatten())\n",
    "        else:  \n",
    "            bookkeeping.append((np.shape(ele)))\n",
    "            vec = np.append(vec, ele.flatten())\n",
    "    bookkeeping.append((Nx, Ny))     \n",
    "    return vec, bookkeeping\n",
    "\n",
    "def vec2coef(vec, bookkeeping):\n",
    "    \"\"\"\n",
    "    Convert an array-type vector to wavelet coefficients, inverse operation of coef2vec.\n",
    "    The initial vector is stocked in a 1D array as follows:\n",
    "    [cAn.flatten(), cHn.flatten(), cVn.flatten(), cDn.flatten(), ..., cH1.flatten(), cV1.flatten(), cD1.flatten()].\n",
    "    After the conversion, the returned wavelet coefficient is in the form of the list as follows:\n",
    "        [cAn, (cHm, cVn, cDn), ..., (cH1, cV1, cD1)],\n",
    "    and each element is a 2D array. This list can be passed as the argument in pywt.waverec2.\n",
    "    \"\"\"\n",
    "    ind = bookkeeping[0][0] * bookkeeping[0][1] \n",
    "    coef = [np.reshape(vec[:ind], bookkeeping[0])]\n",
    "    for ele in bookkeeping[1:-1]:\n",
    "        indnext = ele[0] * ele[1]\n",
    "        coef.append((np.reshape(vec[ind:ind+indnext], ele), \n",
    "                     np.reshape(vec[ind+indnext:ind+2*indnext], ele), \n",
    "                     np.reshape(vec[ind+2*indnext:ind+3*indnext], ele)))\n",
    "        ind += 3*indnext\n",
    "\n",
    "    return coef\n",
    "\n",
    "def wavedec_asarray(im, wv='db8',level=3):\n",
    "    wd = pywt.wavedec2(im,wv,level=level, mode='zero')\n",
    "    wd, book = coef2vec(wd, im.shape[0], im.shape[1])\n",
    "    return wd, book\n",
    "\n",
    "def waverec_asarray(wd, book, wv='db8'):\n",
    "    wc = vec2coef(wd, book)\n",
    "    im = pywt.waverec2(wc,wv, mode='zero')\n",
    "    return im\n",
    "\n",
    "\n",
    "def wavelet_op(im, wav='db8', level=3, test_ajd=False):\n",
    "    \n",
    "    _, b_8 = wavedec_asarray(im,wav,level=level)\n",
    "    #ncoef8 = len(c_8)\n",
    "    \n",
    "    def Psi(x):\n",
    "        out = wavedec_asarray(x,wav,level=level)[0]\n",
    "        return out\n",
    "    \n",
    "    def Psit(y):\n",
    "        out = waverec_asarray(y, b_8, wv=wav)\n",
    "        return out\n",
    "    \n",
    "    if test_ajd is True:\n",
    "        print('-----------------------------')\n",
    "        print('Test wavelet')\n",
    "        xtmp = np.random.rand(*im.shape)\n",
    "        Ptxtmp = Psi(xtmp)\n",
    "        ytmp = np.random.rand(*Ptxtmp.shape)\n",
    "        Pytmp = Psit(ytmp)\n",
    "        fwd = np.sum(xtmp.flatten()*Pytmp.flatten())\n",
    "        bwd = np.sum(Ptxtmp.flatten()*ytmp.flatten())\n",
    "        print('forward: ', fwd)\n",
    "        print('backward: ', bwd)\n",
    "        print('-----------------------------')\n",
    "\n",
    "    return Psi, Psit\n"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.13"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
